﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.IO;
using SyncCover.Properties;

public class Assembly {

    // public static Dictionary<string, string> env = new Dictionary<string, string>();

    readonly string ildasmPath;
    readonly string ilasmPath;

    readonly string assemblyPath;
    readonly string assemblyName;

    readonly string workingFolder;

    readonly string ilFile;
    readonly string ccIlFile;
    readonly string assemblyLibPath;

    List<string> lines;

    public Assembly(string ildasmPath, string ilasmPath, string libPath, string assemblyPath) {
        this.assemblyPath = assemblyPath;
        this.assemblyName = Path.GetFileName(assemblyPath);

        this.ilasmPath = ilasmPath;
        this.ildasmPath = ildasmPath;

        this.workingFolder = Path.Combine(libPath, assemblyName);
        this.assemblyLibPath = Path.Combine(workingFolder, assemblyName);

        this.ilFile = Path.ChangeExtension(assemblyLibPath, "orig.il");
        this.ccIlFile = Path.ChangeExtension(assemblyLibPath, ".il");

        CopyAssemblyAndPdb(assemblyPath, workingFolder);
    }

    public string Name {
        get {
            return assemblyName;
        }
    }

    public FileVersionInfo VersionInfo {
        get {
            return FileVersionInfo.GetVersionInfo(assemblyLibPath);
        }
    }

    static void CopyAssemblyAndPdb(string srcAssembly, string to) {
        string fileName = Path.GetFileName(srcAssembly);

        Directory.CreateDirectory(to);
        File.Copy(srcAssembly, Path.Combine(to, fileName), true);
        File.Copy(Path.ChangeExtension(srcAssembly, ".pdb"), Path.Combine(to, Path.ChangeExtension(fileName, ".pdb")), true);
    }

    public List<string> Lines {
        get {
            if (lines == null) {
                ReadLines();
            }
            return lines;
        }
    }

    void ReadLines() {
        Decompile();
        if (!File.Exists(ilFile)) throw new FileNotFoundException("ilFile", ilFile);

        using (StreamReader sr = new StreamReader(ilFile, Encoding.UTF8)) {
            lines = new List<string>(ReadAllLines(sr));
        }
    }

    public void Decompile() {
        if (!File.Exists(assemblyLibPath)) throw new FileNotFoundException("assemblyLibPath", assemblyLibPath);

        string ildasmArg = string.Format("{0} /out={1} /typelist /linenum /UTF8", assemblyLibPath, ilFile);

        Console.WriteLine("Decompiling: {0}", assemblyName);

        string stdOut, stdErr;
        int exitCode = ProcessUtil.StartAndWaitForResult(ildasmPath, ildasmArg, out stdOut, out stdErr);
        if (exitCode != 0) throw new Exception(stdErr);
    }

    internal void Compile(PeVerify peverify, string outPath) {
        string outAssembly = Path.Combine(outPath, assemblyName);

        bool isDll = Path.GetExtension(assemblyName) == ".dll";
        string ilasmArgs = ccIlFile + (isDll ? " /dll " : " ") + "/OUTPUT=" + outAssembly + " /pdb";

        string stdOut, stdErr;

        Console.WriteLine("Compiling: {0}", assemblyName);
        int exitCode = ProcessUtil.StartAndWaitForResult(ilasmPath, ilasmArgs, out stdOut, out stdErr);
        if (exitCode != 0) throw new Exception(stdErr);

        var expectedResult = peverify.VerifyAssembly(assemblyPath);
        var result = peverify.VerifyAssembly(outAssembly);

        expectedResult.Errors.ForEach(e => result.Errors.Remove(e));

        Console.WriteLine();
        result.Errors.ForEach(a => Console.WriteLine(a));
        Console.WriteLine();

        File.WriteAllLines("result.txt", result.Errors.ToArray());
        File.WriteAllLines("expectedResult.txt", expectedResult.Errors.ToArray());
    }

    static IEnumerable<string> ReadAllLines(StreamReader sr) {
        while (!sr.EndOfStream) {
            yield return sr.ReadLine();
        }
    }

    internal void WriteIL() {
        using (StreamWriter sw = new StreamWriter(ccIlFile)) {
            Lines.ForEach(l => sw.WriteLine(l));
        }
    }

} // class
