﻿using System.Collections.Generic;
using System.IO;
using System;
using System.Diagnostics;
using System.Text;
using System.Globalization;

public class SyncPoint {

    public int SyncId { get; set; }
    public int LineNumber { get; set; }

} // class

public class CoverMapRow {

    public string FileName { get; set; }
    public int SyncId { get; set; }
    public int LineNumber { get; set; }

    public string ClassName { get; set; }
    public string MethodName { get; set; }

    public string Tag { get; set; }

} // class

public class CoverData {

    public int Id;
    public int LineNumber;
    public string FileName;
    public string ClassName;
    public string MethodName;

    public CoverData(int id, int lineNumber, string fileName, string className, string methodName) {
        this.Id = id;
        this.LineNumber = lineNumber;
        this.FileName = fileName;
        this.ClassName = className;
        this.MethodName = methodName;
    }

    public override string ToString() {
        return string.Format("{0}|{1}|{2}|{3}|{4}", Id, FileName, LineNumber, ClassName, MethodName);
    }

} // class

public class CoverMap {

    public static void ToFile(string toFile, string coverTag, IEnumerable<CoverData> data) {
        using (StreamWriter w = new StreamWriter(toFile)) {
            w.WriteLine("T|" + coverTag);
            w.WriteLine("SP|");
            foreach (var d in data) {
                w.WriteLine(d);
            }
        }
    }

    public static IEnumerable<CoverMapRow> FromFile(string fileName) {
        using (StreamReader sr = new StreamReader(fileName)) {

            string versionTag = CoverRun.ExtractField(sr.ReadLine(), "T");
            CoverRun.EatUntil(sr, "SP");

            while (!sr.EndOfStream) {
                string[] parts = sr.ReadLine().Split('|');
                Debug.Assert(parts.Length == 5);

                string theTag = CreateTagFromFileName(parts[1]);

                yield return new CoverMapRow() {
                    SyncId = int.Parse(parts[0]),
                    FileName = parts[1],
                    LineNumber = int.Parse(parts[2]),
                    ClassName = parts[3],
                    MethodName = parts[4],
                    Tag = theTag,
                };
            }
        }
    }

    private static string CreateTagFromFileName(string p) {
        return p;
    }

    public static string GetFileNameFromTag(string versionTag) {
        return Path.Combine(Environment.CurrentDirectory, string.Format("cc.map.{0}.txt", versionTag));
    }

} // class

public enum Coverage {
    None = 0,
    Statement = 1,
    Sync = 2,
}

public class RunData {
    public int SyncId { get; set; }
    public long Hit { get; set; }
    public long Contend { get; set; }

    public bool DidWait {
        get {
            return Contend != 0;
        }
    }

    public bool WasEntered {
        get {
            return Hit != 0 || Contend != 0;
        }
    }

    public Coverage Coverage {
        get {
            if (DidWait) return Coverage.Sync;
            if (WasEntered) return Coverage.Statement;
            return Coverage.None;
        }
    }
}

public class CoverRun {

    public string Program { get; set; }
    public string CommandLine { get; set; }

    public DateTime LogTime { get; set; }
    public string VersionTag { get; set; }
    public IList<RunData> Data { get; set; }

    public string MachineName { get; set; }
    public string OS { get; set; }
    public int ProcCount { get; set; }

    public static IEnumerable<CoverRun> FromFile(string fileName) {

        using (StreamReader sr = new StreamReader(fileName)) {
            while (!sr.EndOfStream) {
                string line = sr.ReadLine();
                if (line == "~|B|~") {
                    throw new Exception("Old Version!");
                } else if (line == "~|B2|~") {
                    throw new Exception("Old Version!");
                } else if (line == "~|B3|~") {
                    CoverRun result = null;
                    try {
                        result = ParseV3(sr);
                    } catch {
                        while (sr.ReadLine() != "~|E|~") { }
                    }

                    if (result != null) {
                        yield return result;
                    }
                }
            }
        }
    }

    static CoverRun ParseV3(StreamReader sr) {
        CoverRun coverRun = new CoverRun();
        coverRun.Program = CoverRun.ExtractField(sr.ReadLine(), "A");
        coverRun.CommandLine = CoverRun.ExtractField(sr.ReadLine(), "C");
        coverRun.LogTime = DateTime.ParseExact(CoverRun.ExtractField(sr.ReadLine(), "D"), "yyyyMMdd`HHmmss", CultureInfo.InvariantCulture);
        coverRun.VersionTag = CoverRun.ExtractField(sr.ReadLine(), "T");

        coverRun.MachineName = CoverRun.ExtractField(sr.ReadLine(), "M");
        coverRun.OS = CoverRun.ExtractField(sr.ReadLine(), "O");
        coverRun.ProcCount = CoverRun.ExtractField<int>(sr.ReadLine(), "P");

        coverRun.Data = new List<RunData>();

        CoverRun.EatUntil(sr, "SP");

        string line;
        while ((line = sr.ReadLine()) != "~|E|~") {
            // 548|Entered
            string[] parts = line.Split('|');
            Debug.Assert(parts.Length == 3);

            coverRun.Data.Add(
                new RunData() {
                    SyncId = int.Parse(parts[0]),
                    Hit = long.Parse(parts[1]),
                    Contend = long.Parse(parts[2]),
                }
            );
        }
        return coverRun;

    }

    public static T1 ExtractField<T1>(string line, string fieldName) {
        string asString = ExtractField(line, fieldName);
        return (T1)Convert.ChangeType(asString, typeof(T1));
    }

    public static void EatUntil(StreamReader sr, string p) {
        while (sr.ReadLine() != p + "|") { }
    }

    public static string ExtractField(string line, string fieldName) {
        string[] parts = line.Split(new[] { '|' }, 2);

        if (parts.Length != 2) {
            throw new Exception(string.Format("ExtractField invalid line={0}", line));
        } else if (parts[0] != fieldName) {
            throw new Exception(string.Format("ExtractField expected={0} actual={1}", fieldName, parts[0]));
        }

        return parts[1];
    }

} // class
